# Function which involve parsing user input
import logging
import re

#REGEX_SS is a global variable and hence all functions have access to it.
REGEX_SS = {
    'S-E-': r"S[0-9][0-9]-[0-9][0-9]E[0-9][0-9]-[0-9][0-9]" , # SXX-YYEXX-YY Range in both Episode and Season
    'SE-' : r"S[0-9][0-9]E[0-9][0-9]-[0-9][0-9]"            , # SXXEXX-YY Season Fixed and Range in Episode
    'S-E' : r"S[0-9][0-9]-[0-9][0-9]E[0-9][0-9]"            , # SXX-YYEYY Episode Fixed and Range in Season
    'S-'  : r"S[0-9][0-9]-[0-9][0-9]"                       , # SXX-YY Range in Season
    'SE'  : r"S[0-9][0-9]E[0-9][0-9]"                       , # SXXEYY Fixed Season Fixed Episode
    'S'   : r"S[0-9][0-9]"                                  , # SXX Fixed Season All Episodes
}

def createRange(subselection:str, excluded=True)-> List[str]:
    """This function converts subselections with a given format into an episode list.
    
    Arguments:
        subselection {str} -- This represents the range of episodes under consideration
    
    Keyword Arguments:
        excluded {bool} -- If True, the episodes will be excluded else they will be included (default: {True})
    
    Returns:
        List[str] -- A List containing all episodes in the range
    """
    # Default values for min and max are 0 and 99. 
    # These are the minimum and maximum digits that can be represented in 2 digits.
    season_min = 0
    season_max = 99
    episode_min = 0
    episode_max = 99
    
    if subselection == '' and excluded :
        return []
    subselection_format = infer_format(subselection)
    
    if (subselection_format.startswith("S-")):
        season_min = int(subselection[1:3])
        season_max = int(subselection[4:6])
    elif (subselection_format.startswith("S")):
        season_min = int(subselection[1:3])
        season_max = season_min
    
    if (subselection_format.endswith("E-")):
        episode_min = int(subselection[-5:-3])
        episode_max = int(subselection[-2:])
    elif (subselection_format.endswith("E")):
        episode_min = int(subselection[-2:])
        season_max = season_min
    
    logging.debug(f"The theoretical range for {subselection} is S{season_min:02d}-{season_max:02d}E{episode_min:02d}-{episode_max:02d}")
    #We want to include both endpoints.
    episodeRange = [f"S{season:02d}E{episode:02d}" for episode in range(episode_min,episode_max+1) for season in range(season_min,season_max+1)]
    logging.debug(f"The episodes in the subselection {subselection} were {episodeRange}")    
    return episodeRange

def infer_format(subselection:str) -> str:
    """
    Function to Infer the Format of User Selection Input and return the REGEX_SS Key to it.
    This function uses regex from Python's re library to find format of the query.
    They are checked in the order of insertion. 
    This is to make sure range selections don't get mixed with singleton selection.
    
    Arguments:
        subselection {str} -- This represents the episode range whose format we need to infer.

    Raises:
        ValueError: Exception is raised if no Subselection ID can be inferred.
    
    Returns:
        str -- Key of the Subselection Format inferred from the Subselection
    """
    for subselection_format in REGEX_SS.keys(): #iterating over REGEX_SS's keys.
        logging.debug(f"Matching {subselection} with {subselection_format} .")
        match = re.search (REGEX_SS[subselection_format], subselection)
        # In re.search returns a None object if it cannot find any match. 
        # However if a match is found we return.
        if (match != None):
            logging.debug(f"Match found. Subselection format is {subselection_format}.")
            return subselection_format
        
    if(subselection_format == "" and subselection != ""):
        logging.exception(f"Illegel Selection from User, Format could not be inferred. Raising Exception")
        raise ValueError(f"Subselection ID Format could be inferred.")
        # Logging the fact that no format is inferred.
    logging.warning(f"Returning Empty String for subselection: {subselection}")
    return "" #Returning an Empty String in case of no format being found.

def processConsiderations(considered_part:str, selected=True) -> List[str]:
    """This function takes in considered_part and returns the range corresponding to it.
    
    Arguments:
        considered_part {str} -- This represents the episodes we are considering for either selection or exclusion.
    
    Keyword Arguments:
        selected {bool} -- Boolean representing the nature of consideration. (default: {True})
    
    Returns:
        List[str] -- all the episodes specified in the considered_part.
    """
    Considered = {True:'Selected', False:'Excluded'}[selected]
    considerationRange = []
    
    logging.info(f"The {Considered} Part is {considered_part}")
    subconsiderations = considered_part.replace(' ','').split(',')
    
    for subconsideration in subconsiderations:
        logging.debug(f"Started processing {subconsideration}.")
        subconsideration_range = createRange(subconsideration, excluded=not selected)
        considerationRange += subconsideration_range #Concat Consideration Lists
        logging.debug(f"Added {subconsideration_range} to {Considered} Files.")
    return considerationRange

def parseSelection(selection:str)-> set[str]:
    """This is the parser function for the selection.
    
    Arguments:
        selection {str} -- String containing both the episodes which are selected and excluded.
    
    Returns:
        set[str] -- A set of all episodes which were specified by the selection string.
    """
    selected_part = ""
    excluded_part = ""
    pos_seperator = selection.find('/')
    if(pos_seperator != -1):
        logging.debug("Seperator Found. Range into Selected and Excluded Part")
        selected_part = selection[:pos_seperator]
        excluded_part = selection[pos_seperator+1:]
    else:
        selected_part = selection

    selected_files = processConsiderations(selected_part,selected=True)
    excluded_files = processConsiderations(excluded_part,selected=False)
    
    #This converts sorted and excluded lists into sets and computes their difference set.               
    logging.debug(f"Removing Excluded files from Selected Files.")
    return set(selected_files)-set(excluded_files) 


#!/usr/bin/env python3

from TVShow import TVShow

def main():
    """The main function. It involves creating TVShow class and running rename method on it.
    """
    tvshow = TVShow()
    tvshow.rename()

# This is the driver code where we call the main function.
if __name__ == "__main__":
    main() 



# importing logging to log and debug the code
# requests,bs4 for webscraping data from TVDB

import logging
import requests
import os
from bs4 import BeautifulSoup
from Episode import Episode
from Rename import renameShow
from UserInput import userInput
from Subtitle import embedSubtitles

class TVShow:
    """
    This is the TV Show class. It is the main class for doing stuff with a TV Show.
    """

    def __init__(self):
        """TV Show Initializer
        The showname may need to include the year it first aired in case there are multiple shows with that name.
        I use userInput() function to populate the parameters.
        """
        user_parameters = userInput()
        self.name = user_parameters['Show Name']
        self.directory = user_parameters['Directory']
        self.output_format = user_parameters['Format']
        self.show_url = ""
        self.titles = [['' for j in range(100)] for i in range(100)]
        self.episodes = {}
        logging.debug(f"New TV Show of {self.name} created from folder - {self.directory}")
        self.find_titles()
    
    def find_titles(self):
        """
        This function find the titles for all episodes of a tvshow at once.
        It iterates from 1 to 99 and stores the title dict in titles.
        Whenever it returns an empty season dict, it means there are no more seasons and we stop.
        The search terms is name of the show with terminal spaces removed and intermediate spaces replaced by -
        The URL for the show The Flash 2014 is "https://www.thetvdb.com/series/The-Flash-2014/"
        
        """
        web_host = "https://www.thetvdb.com/series/"
        search_term = (self.name).strip().replace(" ","-")
        self.show_url = f"{web_host}{search_term}"
        
        for i in range(1,100):
            self.find_season_titles(i)
            if self.titles[i][1] == '': #Implies Season returned with no titles
                logging.error(f"Titles not found for Season {i} of {self.name}")
                break        
    
    def find_season_titles(self,season:int):
        """
        This is the main scraping function that I use. The purpose of this function is to find the titles 
        It uses the BeautifulSoup and requests library.
        I use the TVDB's series database for scraping info.
        This webpage is retrieved by using the request.get() and we build a parsetree using BeautifulSoup()
        We use the find function to search throught the parse tree. We search for the table tag and then table body.
        We will store the episode titles in a dict indexed by the episode number.
        findAll() finds all the matches with a given tag inside the tree. 
        
        Arguments:
            season {int} -- The season # whose TV Show needs to be found. 
        """
        logging.debug(f"Constructing URL of {self.name} Season {season} .")
        web_page_url = f"{self.show_url}/seasons/{str(season)}"
        logging.debug(f"Looking up title information at the URL {web_page_url}")
        
        web_page = requests.get(web_page_url)
        try:
            if(web_page is None):
                logging.error(f"Web Page : {web_page_url} Not Found")
            else:
                logging.debug(f"Web Page : {web_page_url} Found. Parsing content")
                soup = BeautifulSoup(web_page.content, 'html5lib')
                table = soup.find('table', attrs = {'id':'translations'})
                if(table is None):
                    logging.error("Table Not Present")
                else:
                    body = table.find('tbody') #The body of the table
                    if(body is None):
                        logging.error("Body Not Present")
                    else:
                        for row in body.findAll('tr'): #The rows
                            episode_num = row.find('a') #First row is index i.e. episode number.
                            episode_name = row.find('span', attrs = {'data-language':'en'}) #To retrieve english version of the Title
                            if((episode_name is None) and (episode_num is None)):
                                logging.error("Incomplete Information in Rows")
                            else:
                                ep_num = int(episode_num.text.strip())
                                ep_title = episode_name.text.strip()
                                self.titles[season][ep_num] = ep_title #Storing entry in dict.
                                logging.debug(f"{ep_num} -> {ep_title}")
        except Exception as e:
            logging.exception(f"season {season} raised {repr(e)}")
        
    def rename(self):
        """This is the master function for scanning the Show's Folder and then Renaming it.
        We use os.walk() as it gives us all files and subfolders inside the main folder.
        """
        for root,dirs,files in os.walk(self.directory):
            for name in files:
                path_name = os.path.join(root, name)
                try:
                    # We create an instance of the episode class. 
                    episode = Episode(self.name,path_name,self.directory,self.output_format)
                    if episode.isVideo:
                        episode.episode_title = self.titles[episode.season][episode.episode]
                        episode_info = episode.getRenameDetails()
                        # We only add it to episodes if Path Name has actually changed
                        if episode_info['old']['Path Name']!=episode_info['new']['Path Name']:
                            self.episodes[episode.ID()] = episode_info
                
                # Primitive Exception Handling. Most Errors are unavoidable but a response to User is necessary.
                except Exception as e:
                    logging.exception(f"{path_name} raised {repr(e)}")
        renameShow(self.episodes,self.directory)
    
    def embedShowSubtitles(self):
        """The function embeds Subtitles for all files in the directory.
        os.walk() gives us all files and subfolders inside the main folder and then we call embedSubtitles on them
        """
        
        logging.info(f"Checking the directory : {self.show_directory}")
        for root,__,files in os.walk(self.show_directory):
            for name in files:
                path_name = os.path.join(root, name)
                embedSubtitles(path_name,rm_orig=True)


  
# importing logging to log and debug the code
# Importing re for Regex,
import logging
import os
import re

class Episode:
    REGEX = {
        'SE': r"S[0-9][0-9]E[0-9][0-9]", # S00E00 format , most popular and widely used,
        'se': r"s[0-9][0-9]e[0-9][0-9]", # Lowercase version of above format. 
        'S.E':r"S[0-9]+.E[0-9]+", # S0.E0 format , very rare format,
        'SE1':r"S[0-9]+E[0-9]+", # S0E0 format , very rare format,
        'x' : r"[0-9]+x[0-9][0-9]",      # 0x00 format, old school but popular and more compact,
        'X' : r"[0-9]+X[0-9][0-9]",      # lowercase version of above.
        '0' : r"[0-9][0-9][0-9][0-9]*",  # Very poor format which is basically 3 or 4 digits in which last two are for ep# and first/first two are season#
        'v' : r"season[\s]*[0-9]+[\s]*episode[\s]*[0-9]+", # Super verbose format, found knowhere but I use it so I have it.
        'V' : r"Season[\s]*[0-9]+[\s]*[-]*[\s]*Episode[\s]*[0-9]+", # Same thing as above , rare format with a " - " between data
        '2' : r"[0-9][0-9]*",            # This format is bad as we are unaware about episode data. It can be used to find ep# but season# has to inferred from folder names.
    }
    VIDEO_EXTENSIONS = {'.mkv', '.mp4', '.avi', '.srt', '.flv'}
    
    def __init__(self,show_name:str,path_name:str,show_folder:str,output_format:str):
        """Class Initializer
        We infer the ID format of the file and then, we parse the episode and season numbers.
        
        Arguments:
            show_name {str} -- The name of the TV Show
            path_name {str} -- The PATH of the episode.
            show_folder {str} -- The PATH of the TV Show directory.
            output_format {str} -- The output format for the TV Show's ID
        """
        self.show_name = show_name 
        self.season = 0
        self.episode = 0
        self.episode_title = ''
        self.extension = ''
        self.input_format = ''
        self.output_format = output_format
        break_index = path_name.rfind(os.sep) 
        self.episode_folder = path_name[:break_index]
        self.file_name = path_name[break_index+1:]
        self.show_folder = show_folder
        self.isVideo = True
        self.renameDetails = {}
        logging.debug(f"New Episode of {self.show_name} created with filename - {self.file_name}")
        self.parse()
        
    
    def ID(self, output_format:str = "SE",extension=True)-> str:
        """ Function to return ID of episodes in certain format
        
        Keyword Arguments:
            output_format {str} --  (default: {"SE"})
            extension {bool} -- To include the extension or not. (default: {True})
        
        Returns:
            str -- The ID of the episode.
        """
        season_prefix = "S"
        episode_prefix = "E"
        
        if (output_format == "x" or output_format == "X"):
            season_prefix = ""
            episode_prefix = output_format
        elif (output_format == "v"):
            season_prefix = "season "
            episode_prefix = " episode "
        elif (output_format == "V"):
            season_prefix = "Season "
            episode_prefix = " - Episode "
        elif (output_format == "0"):
            season_prefix = ""
            episode_prefix = ""
        
        # We use f-string formatting to extend Episode Number from 1 to 01. 
        season_id = f"{season_prefix}{self.season:02d}"
        episode_id = f"{episode_prefix}{self.episode:02d}"
        
        # Combining Season ID and Episode ID to get ID
        identifier = f"{season_id}{episode_id}"
        if extension:
            identifier += self.extension 
        return identifier
        
    def infer_format(self):
        """
        Function uses regex from Python's re library to find format of the filename.
        The most popular format is SE/se so we check for that first.
        Next we check for 0x00 formats which are also common
        3/4 digit letters are checked next and finally 2 digit letters are checked.
        This is because 2 letter formats might trigger on a 3 digit/4 digit formats so 
        they have to be checked last.
        
        Raises:
            ValueError: Raised in case no ID Format can be inferred.
        """
        logging.debug("Inferring ID Format.")
        for id_format in self.REGEX.keys():
            logging.debug(f"Matching episode with {id_format} .")
            match = re.search (self.REGEX[id_format], self.file_name)
            # In re.search returns a None object if it cannot find any match. 
            # However if a match is found we store the format.
            if (match != None):
                self.input_format = id_format
                logging.debug(f"Match found. Episode's format is {id_format}.")
                break

        if(id_format == ""):
            logging.exception(f"Format could not be inferred. Raising Exception")
            raise ValueError(f"No ID Format could be inferred.")
        # Logging the fact that no format is inferred.
                
    def parse(self):
        """The parse() function retrieves the file extension and episode metadata 
        (season number and episode number which is extracted as string and processed by the Decrypt ID function)     
        """
        logging.debug(f"Parsing the episode.")
        self.infer_format()
        logging.debug("Extracting Episode ID")
        extracted_id = re.search (self.REGEX[self.input_format], self.file_name).group()
        logging.debug("Extracting Episode Extension")
        self.extension = re.search(r".[a-zA-Z|0-9]+\Z",self.file_name).group()
        if self.extension not in self.VIDEO_EXTENSIONS:
            logging.error(f"{self.file_name} isn't a video or subtitle file")
            self.isVideo = False
        logging.debug("Extracting #Season and #Episode")
        self.decryptID(extracted_id)

    def decryptID(self, extracted_id:str):
        """This function extracts the episode and season info from the string supplied to it.
    
        Arguments:
            extracted_id {str} -- This is the extracted string.
        """
        
        logging.debug("Extracting Episode Number.")
        if (self.input_format == 'S.E' or self.input_format == 'SE1') :
            self.episode = int((extracted_id[extracted_id.find('E')+1:]).strip())
        else:
            self.episode = int((extracted_id[-2:]).strip()) # Extract last two letters. This is pretty common.

        logging.debug("Extracting Season Number, if available.")
        if (self.input_format == "SE" or self.input_format == "se" ):
            self.season = int((extracted_id[1:3])) # Extract two letters after S
        elif (self.input_format == "x" or self.input_format == "X" ):
            self.season = int((extracted_id[:-3])) # Extract all letters except last 3. i.e just before x
        elif (self.input_format == "0"):
            self.season = int((extracted_id[:-2])) # Extract all letters except last 3.
        elif (self.input_format == "v" or self.input_format == "V"):
            self.season = int((extracted_id[7:8]).strip()) # Get the letters after "season ". These may include a space if number is single digit so we use strip()
        elif (self.input_format == 'S.E'):
            self.season = int((extracted_id[1:extracted_id.find('.')])) # Extract letters after S and before .
        elif (self.input_format == 'SE1'):
            self.season = int((extracted_id[1:extracted_id.find('E')])) # Extract letters after S and before .
        else:
            logging.debug(f"Season not mentioned in episode with filename - {self.file_name}, infering from {self.episode_folder} .")
            self.season = int(self.episode_folder[-1])
                    
    def cleanName(self,file_name:str) -> str:
        """The function sanitizes the filename by removing all offending characters
        
        Arguments:
            file_name {str} -- The file name which needs to be cleaned in order to sanitize the characters.
        
        Returns:
            str -- The sanitized file name.
        """
        new_name = file_name.replace("*" , "").replace("?" , "").replace(":" , "").replace(">" , "").replace("<" , "").replace("/" , "").replace("\\" , "").replace("\"" , "")
        if (new_name != file_name):
            logging.debug(f"Cleaned {file_name} to {new_name} to remove special characters.")
        return new_name

    def getRenameDetails(self) -> dict:
        """This function creates the rename details of the episode. These are:
        src is the PATH of the original file which is to be renamed.
        season_folder is new folder that may need to be created in case of bad folder names.
        dst is the PATH of renamed file.
        
        Returns:
            dict -- This is dict representing the rename details of the episode.
        """
        src = f"{self.episode_folder}{os.sep}{self.file_name}"
        season_folder = f"{self.show_folder}{os.sep}Season {self.season}" #new name  
        new_name = self.cleanName(self.print())
        dst = f"{season_folder}{os.sep}{new_name}"
        
        self.renameDetails['old'] = {'Path Name' : src , 'Folder Name' : self.episode_folder , 'File Name' : self.file_name}
        self.renameDetails['new'] = {'Path Name' : dst , 'Folder Name' : season_folder , 'File Name' : new_name}
        self.renameDetails['change'] = f"{self.renameDetails['old']['File Name']} -> {self.renameDetails['new']['File Name']}"
        return self.renameDetails 

    def print(self)-> str:
        """ Print complete description of an episode. 
        We trim the show_name by removing the year using the 4 digit REGEX.
        Example if the episode is S01E01 of "The Flash 2009" called the "Pilot" and file extension is .mkv
        The output will be "The Flash S01E01 - Pilot .mkv"
        
        Returns:
            str -- The new file name of the episode with Show Name, Episode ID and Episode Title.
        """
        trimmed_show_name = self.show_name
        match = re.search (r"[0-9][0-9][0-9][0-9]", self.show_name)
        if (match != None):
            trimmed_show_name = trimmed_show_name.replace(match.group(),"").strip()                
        return f"{trimmed_show_name} {self.ID(self.output_format,extension=False)} - {self.episode_title} {self.extension}"

    # This is same as print.
    def __str__(self)-> str:
        """This is same as print
        
        Returns:
            str -- print title of the file.
        """
        return print(self)


# File contains Functions to rename episodes
import logging
import os
from datetime import datetime
from UserInput import getParams,getConfirmation

def previewRename(episodes:dict):
    """This function gives a preview of the name change of the episodes.
    
    Arguments:
        episodes {dict} -- Dictionary containing the rename details of the episodes.
    """
    for episode_id in episodes.keys():
        episode_info = episodes[episode_id]
        logging.warning(episode_info['change'])

def getRenameEpisodes(episodes:dict) -> dict:
    """This function gets the rename_episodes dict from the user selection set. It calls getParams()
    
    Arguments:
        episodes {dict} -- Dictionary containing the rename details of the episodes
    
    Returns:
        dict -- Dictionary which is contains selected episodes that are to be renamed.
    """
    user_selection_set = getParams()
    #This takes an intersection of existing episodes and the user selection range to give us the episodes that can be renamed.
    rename_episodes = {episodeID:episodes[episodeID] for episodeID in episodes if episodeID[:6] in user_selection_set}
    logging.info(f"User selected the following episodes {rename_episodes.keys()}") 
    return rename_episodes

def executeRename(rename_episodes:dict,folder_name:str):
    """This function actually executes the rename of the episodes. 
    It also logs the changes in a history file.
    
    Arguments:
        rename_episodes {dict} -- Dictionary which is contains selected episodes that are to be renamed.
        folder_name {str} -- Path Name of the Directory containing all the episodes.
    """
    # Creating History File to store rename History.
    history_file = open(f'{folder_name}{os.sep}Rename History {datetime.now().strftime("%Y-%m-%d %H.%M")}.log', "w+",encoding='utf-8')
    #Renaming Episodes.
    logging.debug(f"Executing Rename.")
    for episode_id in rename_episodes:
        episode_info = rename_episodes[episode_id]    
        logging.warning(episode_info['change'])
        season_folder = episode_info['new']['Folder Name']
        try:
            if not os.path.exists(season_folder):
                os.makedirs(season_folder)
                logging.debug(f"New Folder created: {season_folder}")
            if os.path.exists(season_folder):
                os.rename(episode_info['old']['Path Name'],episode_info['new']['Path Name'])
                logging.info(f"Rename Executed: Episode is now {episode_info['new']['File Name']}")
                history_file.write(f"{episode_info['change']}\n")
        except Exception as e:
            logging.exception(f"Rename Failed : Raised {repr(e)}. Episode is still {episode_info['old']['File Name']} ")
            history_file.write(f" Rename Failed for {episode_info['old']['Path Name']} due to {repr(e)} \n")
    history_file.close()                    

def renameShow(episodes:dict,folder_name:str):
    """Master Function which renames the Show's episodes.
    
    Arguments:
        episodes {dict} -- Dictionary containing the rename details of the episodes.
        folder_name {str} -- Path Name of the Directory containing all the episodes.
    """
    
    logging.info(f"Previewing New File Names for All Episodes.") 
    previewRename(episodes)
    rename_episodes = getRenameEpisodes(episodes)
    logging.info(f"Previewing New File Names for Selected Episodes.") 
    previewRename(rename_episodes)
    if getConfirmation(process='Rename') :
        executeRename(rename_episodes,folder_name)